#R code to perform sensitivity analysis on sample dataset (simulated data)
#See: Estimating Cost-Effectiveness from Claims and Registry Data with Measured and Unmeasured Confounders
#by E Handorf, D Heitjan, J Bekelman, and N Mitra
##
#Code author:
#Elizabeth Handorf
#Fox Chase Cancer Center

#For use with exampledata.csv
#t = observed survival time
#fail = indicator for death
#obs_cost = total observed cost
#trt = binary treatment indicator
#X = confounding covariate (assumed to be a combination of many covariates, i.e. log of propensity score )


library(geepack)
library(survival)
library(boot)


#Helper function to calculate correction factors given a distribution, parameters, and observed treatment
corct_factors<-function(trt,u_dist,parms,gamma) {
  if (u_dist == "bin") {
    MuX0<-exp(gamma)*parms[1]+(1-parms[1])
    MuX1<-exp(gamma)*parms[2]+(1-parms[2])
    MuXi<-trt*MuX1+(1-trt)*MuX0
  }
  if (u_dist == "norm") {
    MuX0<-exp(gamma*parms[1]+gamma^2*parms[3]^2/2)
    MuX1<-exp(gamma*parms[2]+gamma^2*parms[4]^2/2)
    MuXi<-trt*MuX1+(1-trt)*MuX0
  }
  if (u_dist == "pois") {
    MuX0<-exp(parms[1]*(exp(gamma)-1))
    MuX1<-exp(parms[2]*(exp(gamma)-1))
    MuXi<-trt*MuX1+(1-trt)*MuX0
  }
  if (u_dist == "gamma") {
    MuX0<-(1-parms[3]*gamma)^(-parms[1])
    MuX1<-(1-parms[4]*gamma)^(-parms[2])
    MuXi<-trt*MuX1+(1-trt)*MuX0
  }
  if (u_dist == "none") {
    MuX0<-1
    MuX1<-1
    MuXi<-1
  }
  return(data.frame(cbind(MuX0,MuX1,MuXi)))
}

#### Function to calculate bootstrap confidence intervals

bootstrap.nmb<-function(data,d,u_dist.c,parms.c,gamma.c,u_dist.e,parms.e,gamma.e,WTP,trt.full,x.full) {
  data.boot<-data[d,]
  
  #calculate kaplan-meier estimates of censoring probability for everyone
  #based on their failure time
  censored<-as.numeric(!data.boot$fail)
  km.cens<-survfit(Surv(data.boot$t,censored)~1)
  
  surv<-km.cens$surv
  
  #calculate inverse weights & combine with original dataset
  ipw.cens<-data.frame(cbind(km.cens$time,surv^-1))
  data.boot2<-merge(data.boot,ipw.cens,by.x="t", by.y="X1",all.x=T,all.y=F)
  
  data.boot2<-data.frame(cbind(data.boot2,c(1:length(trt.full))))
  names(data.boot2)[5]<-"X.boot"
  names(data.boot2)[6]<-"ghat"
  names(data.boot2)[7]<-"id"
  
  #remove any rows where P(censoring)=0
  data.boot3<-data.boot2[!is.infinite(data.boot2$ghat),]
  
  #try to run the cost regression
  coef<-coef(geeglm(obs_cost ~ trt + X.boot, data=data.boot3, family=Gamma(link="log"),
                       weights=ghat, subset=fail==1, id=id))
  
  #get naive predicted values from the regression
  
  alpha.star.c<-as.numeric(coef[1])
  beta.star.c<-as.numeric(coef[2])
  theta.star.c<-as.numeric(coef[3])
  
  pred.cost.1<-exp(alpha.star.c + beta.star.c + theta.star.c*x.full)
  pred.cost.0<-exp(alpha.star.c + theta.star.c*x.full)
  
  #find the appropriate correction factors
  cost.CF<-corct_factors(trt.full,u_dist.c,parms.c,gamma.c)  
  
  #calculate the corrected delta for cost
  cost.delta.corrected<-cost.CF$MuXi*(pred.cost.1/cost.CF$MuX1-pred.cost.0/cost.CF$MuX0)
  
  m.delta.corrected.c<-mean(cost.delta.corrected)
  
  #Accelerated failure time model for survival times
  s1<-survreg(Surv(t,fail) ~ trt + X.boot, dist="weibull",data=data.boot3)
  psi_hat<-1/s1$scale
  
  #calculate corrected delta for effects
  eff.CF<-corct_factors(trt.full,u_dist.e,parms.e,gamma.e*psi_hat)  
  
  #calculate the corrected delta for effects
  
  beta.e.hat<--1*s1$coef[2]
  theta.e.hat<--1*s1$coef[3]
  alpha.e.hat<-exp(-1*s1$coef[1]*psi_hat)
  AF.i.hat<-exp(psi_hat*beta.e.hat*trt.full+psi_hat*theta.e.hat*x.full)
  AF.1.hat<-exp(psi_hat*beta.e.hat+psi_hat*theta.e.hat*x.full)
  AF.0.hat<-exp(psi_hat*theta.e.hat*x.full)
  
  #calculate the true causal effect of the treatment on survival time
  
  eff.delta.corrected<-(eff.CF$MuXi/eff.CF$MuX1)^(-1/psi_hat)*(alpha.e.hat*AF.1.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)-
    (eff.CF$MuXi/eff.CF$MuX0)^(-1/psi_hat)*(alpha.e.hat*AF.0.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)
  
  m.delta.corrected.e<-mean(eff.delta.corrected)
  
  m.nmb.corrected<-mean(eff.delta.corrected*WTP-cost.delta.corrected)
  
  return(c(m.delta.corrected.c, m.delta.corrected.e, m.nmb.corrected))
}


######################################  Start analysis here #######################
### Set up inputs

data<-read.csv("C:\\file_path\\exampledata.csv")

# Set Willingness to pay
WTP=50000 


#####################################################################################
#Main analysis without unmeasured confounding
#####################################################################################

#Inverse-weighting to get cost effects
censored<-as.numeric(!data$fail)
km.cens<-survfit(Surv(data$t,censored)~1)

surv<-km.cens$surv


ipw.cens<-data.frame(cbind(km.cens$time,surv^-1))
data2<-merge(data,ipw.cens,by.x="t", by.y="X1",all.x=TRUE,all.y=FALSE)

data2<-data.frame(cbind(data2,c(1:length(data$t))))
names(data2)[6]<-"ghat"
names(data2)[7]<-"id"

data3<-data2[!is.infinite(data2$ghat),]

#run the cost regression
coef<-coef(geeglm(obs_cost ~ trt + X, data=data3, family=Gamma(link="log"),
                     weights=ghat, subset=fail==1, id=id))

alpha.star.c<-as.numeric(coef[1])
beta.star.c<-as.numeric(coef[2])
theta.star.c<-as.numeric(coef[3])

pred.cost.1<-exp(alpha.star.c + beta.star.c + theta.star.c*data3$X)
pred.cost.0<-exp(alpha.star.c + theta.star.c*data3$X)

###Naive predicted cost difference:
naive.cost.diff<-mean(pred.cost.1-pred.cost.0)

#Accelerated failure time model for survival times
s1<-survreg(Surv(t,fail) ~ trt +X, dist="weibull",data=data3)

#transform results to parameterization used in my models
psi_hat<-1/s1$scale

beta.e.hat<--1*s1$coef[2]
theta.e.hat<--1*s1$coef[3]
alpha.e.hat<-exp(-1*s1$coef[1]*psi_hat)


##### Naive predicted effect difference:
AF.i.hat<-exp(psi_hat*beta.e.hat*data3$trt+psi_hat*theta.e.hat*data3$X)
AF.1.hat<-exp(psi_hat*beta.e.hat+psi_hat*theta.e.hat*data3$X)
AF.0.hat<-exp(psi_hat*theta.e.hat*data3$X)

naive.effect.diff<-mean((alpha.e.hat*AF.1.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)-
  (alpha.e.hat*AF.0.hat)^(-1/psi_hat)*gamma(1+1/psi_hat))

##### Naive NMB  
naive.nmb<-naive.effect.diff*WTP - naive.cost.diff

#Bootstrap to get 95% CIs
bootstrap.res<-boot(data3,bootstrap.nmb, R=500,u_dist.c="none",parms.c=0,gamma.c=0,u_dist.e="none",
                    parms.e=0,gamma.e=0,WTP=WTP,trt.full=data3$trt,x.full=data3$X)

delta.boot.ci.c<-boot.ci(bootstrap.res, type="bca",index=1)
delta.boot.ci.e<-boot.ci(bootstrap.res, type="bca",index=2)
delta.boot.ci.nmb<-boot.ci(bootstrap.res, type="bca",index=3)


#################################################################

### Inputs to fuction run sensitivity

### Note: this MUST be run AFTER the naive regression results are aviable in memory

### trt=binary treatment indicator
### x = covariate value
### u.dist.c = distribution of umeasured confounder for cost
### parms.c = parameters of the cost distribution of U in control and treated groups
### strength.c = regression coefficient for cost
### u.dist.e = distribution of umeasured confounder for effect
### parms.e = parameters of the effect distribution of U in control and treated groups
### strenght.e = regression coefficient for effect

###   valid distributions of unmeasured confounders and required parameter values:
# "bin" (bernoulli)
#   parms=(p0,p1) (probability of 1 under control and treatment, respectively)
# "norm" (normal)
#   parms=(mu0,mu1,sigma0,sigma1) (mean and standard deviation in control and treatment, respectively)
# "pois" (poisson)
#   parms=(lambda0,lambda1) (rates in conrol and treatment, respectively)
# "gamma" (gamma)
#   parms=(k0,k1,theta0,theta1) (shape and scale of gamma in treated and control, respectively)
#   note: gamma parameterized as f(u) = 1/(Gamma(k)*theta^k) * u^(k-1) * exp(-u/theta)
# "none"

run_sensitivity<-function(trt,x,
                          u.dist.c,parms.c,strength.c,u.dist.e,parms.e, strength.e)  {
  #find the appropriate correction factors given the assumed unmeasured confounders
  cost.CF<-corct_factors(trt,u.dist.c,parms.c,strength.c)  
  #calculate corrected effect difference
  eff.CF<-corct_factors(trt,u.dist.e,parms.e,strength.e*psi_hat)  


  #calculate the corrected delta for cost

  cost.delta.corrected<-cost.CF$MuXi*(pred.cost.1/cost.CF$MuX1-pred.cost.0/cost.CF$MuX0)

  m.delta.corrected.c<-mean(cost.delta.corrected)


  eff.delta.corrected<-(eff.CF$MuXi/eff.CF$MuX1)^(-1/psi_hat)*(alpha.e.hat*AF.1.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)-
    (eff.CF$MuXi/eff.CF$MuX0)^(-1/psi_hat)*(alpha.e.hat*AF.0.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)

  #Combine to get net monetary benefits  
  m.delta.corrected.e<-mean(eff.delta.corrected)
  m.nmb.corrected<-m.delta.corrected.e*WTP-m.delta.corrected.c
  
  #Bootstrap to get 95% CIs
  bootstrap.res<-boot(data3,bootstrap.nmb, R=500,u_dist.c=u.dist.c,parms.c=parms.c,gamma.c=strength.c,u_dist.e=u.dist.e,
                      parms.e=parms.e,gamma.e=strength.e,WTP=WTP,trt.full=data3$trt,x.full=data3$X)
  
  delta.boot.ci.c<-boot.ci(bootstrap.res, type="bca",index=1)
  delta.boot.ci.e<-boot.ci(bootstrap.res, type="bca",index=2)
  delta.boot.ci.nmb<-boot.ci(bootstrap.res, type="bca",index=3)
  
  lcl.e<-delta.boot.ci.e$bca[4]
  ucl.e<-delta.boot.ci.e$bca[5]
  lcl.c<-delta.boot.ci.c$bca[4]
  ucl.c<-delta.boot.ci.c$bca[5]
  lcl.nmb<-delta.boot.ci.nmb$bca[4]
  ucl.nmb<-delta.boot.ci.nmb$bca[5]
  
  return(c(Effect=m.delta.corrected.e,LCL.E=lcl.e,UCL.E=ucl.e,
           Cost=m.delta.corrected.c,LCL.C=lcl.c,UCL.C=ucl.c,
           NMB=m.nmb.corrected,LCL.NMB=lcl.nmb,UCL.NMB=ucl.nmb))
}


##########################################
#Summarize naive results (not accounting for an unmeasured confounder)
##########################################
#Effect
print(naive.effect.diff)
print(delta.boot.ci.e$bca[4:5])

#Cost
print(naive.cost.diff)
print(delta.boot.ci.c$bca[4:5])

#NMB
print(naive.nmb)
print(delta.boot.ci.nmb$bca[4:5])

##########################################
#Try different parameter combinations for sensitivity analysis

#E.g. bernoulli confounder for cost, normal confounder for effect
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.25*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.75),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.25*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.75),  strength.c=.5,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.25*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.75),  strength.c=.75,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.25*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.1,0.9),  strength.c=.75,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.25*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.5*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.75*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.75,1,1),  strength.e=.5*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.75,1,1),  strength.e=.75*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.75,1,1),  strength.e=1*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.25,0.5),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.75,1,1),  strength.e=1.25*psi_hat)


####### Conclusion for example dataset/sensitivity analysis:
####### NMB is positive but not significantly different from zero.
####### NMB results are fairly insensitive to even large confounders

####### Other distributions for unmeasured confounders

run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="bin", parms.c = c(0.1,0.9),  strength.c=.75,
                u.dist.e="none", parms.e = 0,  strength.e=0*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="none", parms.c = 0,  strength.c=0,
                u.dist.e="bin", parms.e = c(0.1,0.9),  strength.e=.75*psi_hat)

#Normal
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="norm", parms.c = c(0,0.5,1,1),  strength.c=.25,
                u.dist.e="norm", parms.e = c(0,0.5,1,1),  strength.e=.75*psi_hat)

#Poisson
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="pois", parms.c = c(0.5,1.5),  strength.c=.25,
                u.dist.e="pois", parms.e = c(0.5,1.5),  strength.e=.75*psi_hat)

#Constraint for gamma parameters: strength < (1/theta)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="gamma", parms.c = c(1.1,1,3,3),  strength.c=.1,
                u.dist.e="gamma", parms.e = c(1.1,1,3,3),  strength.e=.1*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="gamma", parms.c = c(1.1,1,3,3),  strength.c=.2,
                u.dist.e="gamma", parms.e = c(1.1,1,3,3),  strength.e=.2*psi_hat)
run_sensitivity(trt=data3$trt, x=data$X,
                u.dist.c="gamma", parms.c = c(1.1,1,3,3),  strength.c=-.5,
                u.dist.e="gamma", parms.e = c(1.1,1,3,3),  strength.e=-.5*psi_hat)





