#R code to evaluate sensitivity analysis procedure using simulations
#Data is simulated, naive model fit, then correction is applied.  95% CIs generated via a bootstrap
#See: Estimating Cost-Effectiveness from Claims and Registry Data with Measured and Unmeasured Confounders
#by E Handorf, D Heitjan, J Bekelman, and N Mitra
##
#Code author:
#Elizabeth Handorf
#Fox Chase Cancer Center


#Note: this function is written to be run in batch (in a linux environment), with the distributions and 
#strengths of unmeasured confounders supplied via the command line.  
#It can be run interactively by changing the default values used if no command line argments are supplied

#m = number of simulations
#2n= total number of subjects 
m<-500
n<-200
set.seed(147852)
library(geepack)
library(survival)
library(boot)
library(copula)
library(batch)

#Helper functions to generate unmeasured confounders

fun.bin.u<-function(parms,n) {
  #generate unknown confounder
  #bernoulli distribution
  pi0<-parms[1]
  pi1<-parms[2]
  u0<-rbinom(n,1,pi0)
  u1<-rbinom(n,1,pi1)
  u<-c(u0,u1)
  return(u=u)
}

fun.norm.u<-function(parms,n) {
  #generate unknown confounder
  #normal distribution
  mu0<-parms[1]
  mu1<-parms[2]
  sigma<-1
  u0<-rnorm(n,mean=mu0,sd=sigma)
  u1<-rnorm(n,mean=mu1,sd=sigma)
  u<-c(u0,u1)
}
fun.pois.u<-function(parms,n) {
  #generate unknown confounder
  #poisson distribution
  lambda0<-parms[1]
  lambda1<-parms[2]
  u0<-rpois(n,lambda0)
  u1<-rpois(n,lambda1)
  u<-c(u0,u1)
}
fun.gamma.u<-function(parms,n) {
  #generate unknown confounder
  #gamma distribution
  shape0<-parms[1]
  shape1<-parms[2]
  scale0<-parms[3]
  scale1<-parms[4]
  u0<-rgamma(n,shape0,scale=scale0)
  u1<-rgamma(n,shape1,scale=scale1)
  u<-c(u0,u1)
}

#helper function to calculate correction factors given a distribution, parameters, and observed treatment
corct_factors<-function(trt,u_dist,parms,gamma) {
    if (u_dist == "bin") {
      MuX0<-exp(gamma)*parms[1]+(1-parms[1])
      MuX1<-exp(gamma)*parms[2]+(1-parms[2])
      MuXi<-trt*MuX1+(1-trt)*MuX0
    }
    if (u_dist == "norm") {
      MuX0<-exp(gamma*parms[1]+gamma^2*parms[3]^2/2)
      MuX1<-exp(gamma*parms[2]+gamma^2*parms[4]^2/2)
      MuXi<-trt*MuX1+(1-trt)*MuX0
    }
    if (u_dist == "pois") {
      MuX0<-exp(parms[1]*(exp(gamma)-1))
      MuX1<-exp(parms[2]*(exp(gamma)-1))
      MuXi<-trt*MuX1+(1-trt)*MuX0
    }
    if (u_dist == "gamma") {
      MuX0<-(1-parms[3]*gamma)^(-parms[1])
      MuX1<-(1-parms[4]*gamma)^(-parms[2])
      MuXi<-trt*MuX1+(1-trt)*MuX0
    }
    if (u_dist == "none") {
      MuX0<-1
      MuX1<-1
      MuXi<-1
    }
    return(data.frame(cbind(MuX0,MuX1,MuXi)))
}

#####################################################################################################
##Boostrapping function to calculate mean costs, effects, NMB
##Obtain distribution of regression coefficients using non-parametric boostrap samples
##Then calculate fitted means using the original dataset (condition on X's)
#####################################################################################################
calc.mean.nmb<-function(data,d,u_dist.c,parms.c,gamma.c,u_dist.e,parms.e,gamma.e,WTP,trt.full,x.full) {
  data.boot<-data[d,]

  #calculate kaplan-meier estimates of censoring probability for everyone
  #based on their failure time
  censored<-as.numeric(!data.boot$fail)
  km.cens<-survfit(Surv(data.boot$t,censored)~1)

  surv<-km.cens$surv

  #calculate inverse weights & combine with original dataset
  ipw.cens<-data.frame(cbind(km.cens$time,surv^-1))
  data2<-merge(data.boot,ipw.cens,by.x="t", by.y="X1",all.x=T,all.y=F)

  data2<-data.frame(cbind(data2,c(1:(2*n))))
  names(data2)[11]<-"ghat"
  names(data2)[12]<-"id"

  #remove any rows where P(censoring)=0
  data3<-data2[!is.infinite(data2$ghat),]

 #try to run the cost regression
  coef<-tryCatch(
      model<-coef(geeglm(obs_cost ~ trt + x.c, data=data3, family=Gamma(link="log"),
        weights=ghat, subset=fail==1, id=id)),
      error=function(e) NA
  )

  #get naive predicted values from the regression
  
  alpha.star.c<-as.numeric(coef[1])
  beta.star.c<-as.numeric(coef[2])
  theta.star.c<-as.numeric(coef[3])

  pred.cost.1<-exp(alpha.star.c + beta.star.c + theta.star.c*x.full)
  pred.cost.0<-exp(alpha.star.c + theta.star.c*x.full)

  #find the appropriate correction factors
  cost.CF<-corct_factors(trt.full,u_dist.c,parms.c,gamma.c)  

  #calculate the corrected delta for cost
  cost.delta.corrected<-cost.CF$MuXi*(pred.cost.1/cost.CF$MuX1-pred.cost.0/cost.CF$MuX0)

  m.delta.corrected.c<-mean(cost.delta.corrected)

  #Accelerated failure time model for survival times
  s1<-survreg(Surv(t,fail) ~ trt + x.e, dist="weibull",data=data2)
    psi_hat<-1/s1$scale

  #calculate corrected delta for effects
  eff.CF<-corct_factors(trt.full,u_dist.e,parms.e,gamma.e*psi_hat)  

  #calculate the corrected delta for effects
  
  beta.e.hat<--1*s1$coef[2]
  theta.e.hat<--1*s1$coef[3]
  alpha.e.hat<-exp(-1*s1$coef[1]*psi_hat)
  AF.i.hat<-exp(psi_hat*beta.e.hat*trt.full+psi_hat*theta.e.hat*x.full)
  AF.1.hat<-exp(psi_hat*beta.e.hat+psi_hat*theta.e.hat*x.full)
  AF.0.hat<-exp(psi_hat*theta.e.hat*x.full)

  #calculate the true causal effect of the treatment on survival time
 
  eff.delta.corrected<-(eff.CF$MuXi/eff.CF$MuX1)^(-1/psi_hat)*(alpha.e.hat*AF.1.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)-
   (eff.CF$MuXi/eff.CF$MuX0)^(-1/psi_hat)*(alpha.e.hat*AF.0.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)

  m.delta.corrected.e<-mean(eff.delta.corrected)

  m.nmb.corrected<-mean(eff.delta.corrected*WTP-cost.delta.corrected)

  return(c(m.delta.corrected.c, m.delta.corrected.e, m.nmb.corrected))
}

#####################################################################################################
######## Main function to generate simulated data, and test the correction for unmeasured confounding
#####################################################################################################

fun.sim.gamma_reg<-function(gamma.c,gamma.e,parms.c,parms.e,u_dist.c,u_dist.e,WTP) {
 #m<-50
 true.covered.cost<-rep(NA,m)
 pct.bias.cost<-rep(NA,m)
 true.covered.eff<-rep(NA,m)
 pct.bias.eff<-rep(NA,m)
 true.covered.nmb<-rep(NA,m)
 pct.bias.nmb<-rep(NA,m)

 cor.cost.eff<-rep(NA,m)
 pr.cens<-rep(NA,m)
 
 #monitor how many itterations successfully complete
 reg.success<-rep(1,m)
 ci.success<-rep(1,m)

 for (j in 1:m) {


  #treatment
  trt<-c(rep(0,n),rep(1,n))
  

  #unmeasured confounder for costs
  if (u_dist.c == "bin") {
    u.c<-fun.bin.u(parms=parms.c,n=n)
  } else if (u_dist.c == "norm") {
    u.c<-fun.norm.u(parms=parms.c,n=n)
  } else if (u_dist.c == "pois") {
    u.c<-fun.pois.u(parms=parms.c,n=n)
  } else if (u_dist.c == "gamma") {
    u.c<-fun.gamma.u(parms=parms.c,n=n)
  } else if (u_dist.c == "none") {
    u.c<-rep(0,2*n)
  }

  #unmeasured confounder for effects

  #if the unmeasured covariate has the same distribution, use the same confounder
  if (u_dist.e == u_dist.c) {
     u.e<-u.c
  } else if (u_dist.e == "bin") {
    u.e<-fun.bin.u(parms=parms.e,n=n)
  } else if (u_dist.e == "norm") {
    u.e<-fun.norm.u(parms=parms.e,n=n)
  } else if (u_dist.e == "pois") {
    u.e<-fun.pois.u(parms=parms.e,n=n)
  } else if (u_dist.e == "gamma") {
    u.e<-fun.gamma.u(parms=parms.e,n=n)
  } else if (u_dist.e == "none") {
    u.e<-rep(0,2*n)
  }



  #generate measured covariates via propensity score
  x.c<-c(rnorm(n,mean=0,sd=1),rnorm(n,mean=1,sd=1))  

  #make the covariates for costs and effects the same - analogous to the propensity score
  x.e<-x.c

  #generate expected mean for a given covariate combination
  alpha.c<-5
  beta.c<-1
  theta.c<-0.5

  log_e_mean.c<-alpha.c+beta.c*trt+gamma.c*u.c+theta.c*x.c
  e_mean.c<-exp(log_e_mean.c)

  #Now calculate the true causal effect of the treatment on cost
  true_delta.c<-exp(alpha.c+beta.c+gamma.c*u.c+theta.c*x.c)-exp(alpha.c+gamma.c*u.c+theta.c*x.c)
  m.true.delta.c<-mean(true_delta.c)

  #Effects
  psi<-0.5
  alpha.e<-.1
  beta.e<--0.5
  theta.e<--0.25

  AF.true<-exp(psi*beta.e*trt+psi*theta.e*x.e+psi*gamma.e*u.e)
  AF.1<-exp(psi*beta.e+psi*theta.e*x.e+psi*gamma.e*u.e)
  AF.0<-exp(psi*theta.e*x.e+psi*gamma.e*u.e)

  #calculate the true causal effect of the treatment on survival time
  true_delta.e<-(alpha.e*AF.1)^(-1/psi)*gamma(1+1/psi)-(alpha.e*AF.0)^(-1/psi)*gamma(1+1/psi)
  m.true.delta.e<-mean(true_delta.e)

  #Also, calculate the true causal effect on the NMB
  m.true.NMB<-m.true.delta.e*WTP-m.true.delta.c

  #Now, generate correlated costs and survival times 
 
  #for costs, let the scale paramter equal 1.
  
  #Generate correlated costs and survial times using a copula function
  my.copula<-archmCopula(family="clayton",dim=2,param=0.5)
  outcome<-matrix(rep(NA,4*n),nrow=2*n)
  for (i in 1:(2*n)) {
      my.mvd<-mvdc(copula=my.copula, margins = c("gamma","weibull"),
        paramMargins = list(list(shape=e_mean.c[i],scale=1),list(shape=psi,scale=(alpha.e*AF.true[i])^(-1/psi))))
      #outcome[i,]<-rmvdc(my.mvd,1)  #Old version
          outcome[i,]<-rMvdc(1,my.mvd) #New version
  }

  #What is the observed correlation (corrected for the mean) of costs and effects?
  m.c<-outcome[,1]-e_mean.c
  m.e<-outcome[,2]-(alpha.e*AF.true)^(-1/psi)*gamma(1+1/psi)
  cor.cost.eff[j]<-cor(m.c,m.e)

  # Generate uniform censoring time
  cens.t<-runif(2*n,min=0,max=2000)

  #Find the minimum of the failure and censoring times & get censoring indicators
  f<-outcome[,2]
  cost<-outcome[,1]
  input<-cbind(f,cens.t)
  t<-apply(input,1,min)
  fail<-(t==f) 
  

  #censored cost is proportional to the amount of true time observed
  obs_cost<-cost*(t/f)

  data<-data.frame(cbind(t,f,fail,obs_cost,cost,trt,u.e,u.c,x.e,x.c))

  ##### At this point, the data is all generated.  Now, analyze based on the reduced model,
  ##### applying the appropriate correction factors for the unmeasured variables U.e and U.c

  censored<-as.numeric(!data$fail)
  
  pr.cens[j]<-mean(censored)
  
  km.cens<-survfit(Surv(data$t,censored)~1)

  #deal with error caused by censoring probability =0
  surv<-km.cens$surv

  ipw.cens<-data.frame(cbind(km.cens$time,surv^-1))
  data2<-merge(data,ipw.cens,by.x="t", by.y="X1",all.x=TRUE,all.y=FALSE)

  data2<-data.frame(cbind(data2,c(1:(2*n))))
  names(data2)[11]<-"ghat"
  names(data2)[12]<-"id"

  data3<-data2[!is.infinite(data2$ghat),]
  
  #try to run the cost regression
  coef<-tryCatch(
      model<-coef(geeglm(obs_cost ~ trt + x.c, data=data3, family=Gamma(link="log"),
        weights=ghat, subset=fail==1, id=id)),
      error=function(e) NA
  )

  alpha.star.c<-as.numeric(coef[1])
  beta.star.c<-as.numeric(coef[2])
  theta.star.c<-as.numeric(coef[3])

  pred.cost.1<-exp(alpha.star.c + beta.star.c + theta.star.c*data3$x.c)
  pred.cost.0<-exp(alpha.star.c + theta.star.c*data3$x.c)

  #find the appropriate correction factors
  cost.CF<-corct_factors(data3$trt,u_dist.c,parms.c,gamma.c)  

  #calculate the corrected delta for cost
  cost.delta.corrected<-cost.CF$MuXi*(pred.cost.1/cost.CF$MuX1-pred.cost.0/cost.CF$MuX0)

  m.delta.corrected.c<-mean(cost.delta.corrected)

  #Accelerated failure time model for survival times
  s1<-survreg(Surv(t,fail) ~ trt + x.e, dist="weibull",data=data2)
    psi_hat<-1/s1$scale

  #calculate corrected effect difference
  eff.CF<-corct_factors(data2$trt,u_dist.e,parms.e,gamma.e*psi_hat)  
  
  beta.e.hat<--1*s1$coef[2]
  theta.e.hat<--1*s1$coef[3]
  alpha.e.hat<-exp(-1*s1$coef[1]*psi_hat)
  AF.i.hat<-exp(psi_hat*beta.e.hat*data2$trt+psi_hat*theta.e.hat*data2$x.e)
  AF.1.hat<-exp(psi_hat*beta.e.hat+psi_hat*theta.e.hat*data2$x.e)
  AF.0.hat<-exp(psi_hat*theta.e.hat*data2$x.e)

  eff.delta.corrected<-(eff.CF$MuXi/eff.CF$MuX1)^(-1/psi_hat)*(alpha.e.hat*AF.1.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)-
   (eff.CF$MuXi/eff.CF$MuX0)^(-1/psi_hat)*(alpha.e.hat*AF.0.hat)^(-1/psi_hat)*gamma(1+1/psi_hat)

  #Combine to get net monetary benefits  
  m.delta.corrected.e<-mean(eff.delta.corrected)
  m.nmb.corrected<-m.delta.corrected.e*WTP-m.delta.corrected.c

 #Use bootstrap to get 95% CIs.  Use error catching because occasionally the bootstrap fails.

 delta.boot2<-tryCatch(
      boot(data,calc.mean.nmb, R=500,u_dist.c=u_dist.c,parms.c=parms.c,gamma.c=gamma.c,u_dist.e=u_dist.e,
      parms.e=parms.e,gamma.e=gamma.e,WTP=WTP,trt.full=data2$trt,x.full=data2$x.c),
      error=function(e) NA
  )


  delta.boot.ci.c<-tryCatch(boot.ci(delta.boot2, type="bca",index=1),	error=function(e) NA)
  delta.boot.ci.e<-tryCatch(boot.ci(delta.boot2, type="bca",index=2) ,error=function(e) NA)
  delta.boot.ci.nmb<-tryCatch(boot.ci(delta.boot2, type="bca",index=3),error=function(e) NA)
  
  #Extract the confidence intervals from all of these
  lcl.e<-tryCatch(delta.boot.ci.e$bca[4],error=function(e) NA)
  ucl.e<-tryCatch(delta.boot.ci.e$bca[5],error=function(e) NA)
  lcl.c<-tryCatch(delta.boot.ci.c$bca[4],error=function(e) NA)
  ucl.c<-tryCatch(delta.boot.ci.c$bca[5],error=function(e) NA)
  lcl.nmb<-tryCatch(delta.boot.ci.nmb$bca[4],error=function(e) NA)
  ucl.nmb<-tryCatch(delta.boot.ci.nmb$bca[5],error=function(e) NA)

  #summary of coverage and percent bias
  true.covered.cost[j]<-lcl.c<=m.true.delta.c & ucl.c>=m.true.delta.c
  pct.bias.cost[j]<-(m.delta.corrected.c-m.true.delta.c)/m.true.delta.c
 
  true.covered.eff[j]<-lcl.e<=m.true.delta.e & ucl.e>=m.true.delta.e
  pct.bias.eff[j]<-(m.delta.corrected.e-m.true.delta.e)/m.true.delta.e

  true.covered.nmb[j]<-lcl.nmb<=m.true.NMB & ucl.nmb>=m.true.NMB
  pct.bias.nmb[j]<-(m.nmb.corrected-m.true.NMB)/m.true.NMB

  #how many of the regressions & bootstrap approaches were successful?

  if(is.na(coef)) {reg.success[j]<-0}
  if(is.na(lcl.c)) {ci.success[j]<-0}

 } 
 return(data.frame(cbind(true.covered.cost,
	pct.bias.cost,true.covered.eff,pct.bias.eff,
	true.covered.nmb,pct.bias.nmb,cor.cost.eff,
	reg.success,ci.success,pr.cens)))
}


#Defualt values for types and strengths of unmeasured confounders can be changed here
#0=None
#1=Bernoulli
#2=Normal
#3=Poisson
#4=Gamma

#Some code to extract the inputs from the command line, modified from 
#http://yangfeng.wordpress.com/2009/09/03/including-arguments-in-r-cmd-batch-mode/
##First read in the arguments listed at the command line
args=(commandArgs(TRUE))

##args is now a list of character vectors
## First check to see if arguments are passed.
## Then cycle through each element of the list and evaluate the expressions.
if(length(args)==0){
    print("No arguments supplied.")
    ##supply default values

    uC <- 0
    uE <- 0
	gamC <- 0.75
	gamE <- 0.375
}else{
    for(i in 1:length(args)){
         eval(parse(text=args[[i]]))
    }
}


#get parameter values used in simulations based on distributions of the umeasured confounders
 if (uC == 1) {
    uC<-"bin"
    parmsC<-c(0.3,0.866)
  } else if (uC == 2) {
    uC<-"norm"
    parmsC<-c(0,1,1,1)
  } else if (uC == 3) {
    uC<-"pois"
    parmsC<-c(1,1.58)
  } else if (uC == 4) {
    uC<-"gamma"
    parmsC<-c(0.5,0.868,0.75,0.75)
  } else if (uC == 0) {
    uC<-"none"
    parmsC<-0
  }
 if (uE == 1) {
	uE<-"bin"
    parmsE<-c(0.3,0.866)
  } else if (uE == 2) {
	uE<-"norm"
    parmsE<-c(0,1,1,1)
  } else if (uE == 3) {
	uE<-"pois"
    parmsE<-c(1,1.58)
  } else if (uE == 4) {
	uE<-"gamma"
    parmsE<-c(0.5,0.868,0.75,0.75)
  } else if (uE == 0) {
	uE<-"none"
    parmsE<-0
  }


#########################################################################
#Call the main function
stime<-proc.time()[3] #Timer

sim1<-fun.sim.gamma_reg(gamma.c=gamC,gamma.e=gamE,parms.c=parmsC,parms.e=parmsE,
  u_dist.c=uC,u_dist.e=uE,WTP=137)
#output results summary
c(uC,uE,gamC,gamE)
apply(sim1,2,mean,na.rm=T)
apply(sim1,2,mean)
#calculate elapsed time
ftime<-proc.time()[3]
elapsed_time<-(ftime-stime)/60
elapsed_time
